/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: playwav.c,v 1.1.1.1 2002/07/30 11:02:37 zinnia Exp $
 */

#include <stdio.h>
#include <fcntl.h>

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <unistd.h>

#include "playwav.h"
#include "endian.h"
#include "converter.h"
#include "drivers.h"
#include "format.h"

void play_loop(int infd, driver_vector* driver, int loopmax, int wavflag)
{
  static char buf[0x1000];
  int loopcount = 0;
  int len;

  while(loopcount < loopmax){
    len = read(infd, buf, 0x1000);
    if(len == 0){
      loopcount++;
      driver->audio_sync();
      lseek(infd, wavflag * 44, SEEK_SET);
      continue;
    }
    if(len < 0) break;

    driver->audio_play(buf, len);
  }
  driver->audio_sync();
}

int main(int argc, char* argv[])
{
  int infd = 0;
  int wavflag = 0;
  char* filename = NULL;
  char* arg;
  char* formatstring = DEFAULT_PLAY_FORMAT;
  char* drivername = DEFAULT_PLAY_DRIVER;

  int ch = DEFAULT_PLAY_CH;
  int rate = DEFAULT_PLAY_RATE;
  int loopmax = 1;

  int i;
  char buf[44];
  driver_vector* player;
  formatunit* format;

  endian_converter ec;


  for(i = 1; i < argc; i++){
    if(!strcmp(argv[i], "-")){
      infd = 0;
    } else {
#define NEXT_ARG() (*(argv[i] + 2) == '\0' ? argv[++i] : argv[i] + 2)
      if(*argv[i] == '-'){
	switch(*(argv[i] + 1)){
	case 'd':
	  arg = NEXT_ARG();
	  drivername = arg;
	  break;

	case 'f':
	  arg = NEXT_ARG();
	  formatstring = arg;
	  break;

	case 'c':
	  arg = NEXT_ARG();
	  ch = atoi(arg);
	  break;
	case 'r':
	  arg = NEXT_ARG();
	  rate = atoi(arg);
	  break;

	case 'l':
	  arg = NEXT_ARG();
	  loopmax = atoi(arg);
	  break;

	case 'h':
	  /* FALLTHROUGH */
	case '?':
	  /* FALLTHROUGH */
	default:
	  usage();
	  break; /* NOTREACHED */
	}
      } else {
	filename = argv[i];
      }
    }
  }

  if(filename != NULL){
    infd = open(filename, O_RDONLY | O_BINARY);
    if(infd < 0){
      fprintf(stderr, "%s: %s\n", filename, strerror(errno));
      return 1;
    }
  }
  FORCE_BINMODE(infd);

  i = read(infd, buf, 44);
  if(!strncmp(&buf[0], "RIFF", 4) && !strncmp(&buf[8], "WAVE", 4)){
    wavflag = 1;
    if(buf[20] != 1){
      fprintf(stderr, "unexpected WAV format type(%d)\n", buf[21]);
      return 1;
    }
    get_endianconverter(&ec);
    ch = ec.le2sys16(*(short*)(&buf[22]));
    rate = ec.le2sys32(*(int*)(&buf[24]));
    switch(ec.le2sys16(*(short*)(&buf[34]))){
    case 8:
      formatstring = "s8";
      break;
    case 16:
      formatstring = "s16";
      break;
    default:
      fprintf(stderr, "unexpected bit format\n");
      return 1;
      break; /* NOTREACHED */
    }
  }

  player = query_driver(drivername);
  if(player == NULL){
    fprintf(stderr, "driver %s not available.\n", drivername);
    usage();
  }
  format = query_format(formatstring);
  if(format == NULL){
    fprintf(stderr, "format %s not available.\n", formatstring);
    usage();
  }

  /* show status */
  fprintf(stderr, 
	  "filename: %s\n"
	  "type: %s\n"
	  "rate: %d\n"
	  "format: %d bit, %ssigned linear%s\n"
	  "channel: %d\n"
	  "loop: %d\n",
	  
	  filename == NULL ? "(stdin)" : filename,
	  wavflag == 1 ? "WAV" : "RAW(guess)",
	  rate,
	  format->bit, 
	  format->bitformat == UNSIGNED_LINEAR ? "un" : "",
	  format->bit == 8 ? "" : (format->endian == LITTLEENDIAN) ? ", little endian" : ", big endian",
	  ch,
	  loopmax);


  if(format->bit != 8){
    if(format->endian == get_systemendian()){
    } else {

      /* swap vector */
    }
  }

  if(format->bitformat == SIGNED_LINEAR){
  } else {
    /* signature vector */
  }

  if(player->audio_open(format->bit, rate, ch) != 0){
    fprintf(stderr, "can't initialize audio driver.\n");
    return 1;
  }
  if(wavflag == 0){
    player->audio_play(buf, i);
  }
  play_loop(infd, player, loopmax, wavflag);

  player->audio_close();

  return 0;
}
