/*
 * Copyright (c) 2002 by Tsuyoshi Iguchi (zinnia@risky-safety.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: ringbuffer.c,v 1.1.1.1 2002/07/30 11:02:37 zinnia Exp $
 */

/* ringbuffer.c  -- ring buffer manager */
#include "ringbuffer.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define RINGBUFFER_WAIT_RESOLUTION 10

#define RINGBUFFER_DEBUG
#undef RINGBUFFER_DEBUG

#undef DEBUG
#ifdef RINGBUFFER_DEBUG
#define DEBUG(expression) expression
#else
#define DEBUG(expression) /**/
#endif

int ringbuffer_init(ringbuffer_context* context, int s)
{
  context->size = s;
  context->rpos = context->wpos = 0;
  context->inlen = 0;
  context->delayfunction = NULL;
  context->blocking_bitmap = RINGBUFFER_BLOCKING_NONE;

  context->buf = (unsigned char*) malloc(s);
  memset(context->buf, 0, s);

  if(context->buf == NULL) return -1;
  return 0;
}

void ringbuffer_register_blocking_vector(ringbuffer_context* context, void (*dl)(int msec))
{
  context->delayfunction = dl;
}

int ringbuffer_register_blocking(ringbuffer_context* context, int bmode){
  int oldbitmap = context->blocking_bitmap;
  context->blocking_bitmap = bmode;
  return oldbitmap;
}

void ringbuffer_terminate(ringbuffer_context* context)
{
  if(context->buf != NULL) free(context->buf);
}

#define SHOW_STATUS() DEBUG(fprintf(stderr, "result: %d, rpos: %d, wpos: %d, inlen: %d\n", result, context->rpos, context->wpos, context->inlen))

int ringbuffer_get(ringbuffer_context* context, unsigned char* buf, int size)
{
  int result;
  int remain;
  int available_atonce;
  unsigned char* destpointer;

  for(destpointer = buf, result = 0; result < size;){
    DEBUG(fprintf(stderr, "ringbuffer_get entrance(request = %d)\n", size));
    SHOW_STATUS();
    if(context->inlen == 0){  /* buffer empty */
      if((context->blocking_bitmap & RINGBUFFER_BLOCKING_GET) &&
	 context->delayfunction != NULL){
	DEBUG(fprintf(stderr, "buffer empty. waiting...\n"));
	context->delayfunction(RINGBUFFER_WAIT_RESOLUTION);
      } else {
	break;
      }
    } else {
      remain = size - result;
      if(context->rpos < context->wpos){
	/*
	  0-------R------------W-------size (inlen = W - R)
                  =============
	*/
	available_atonce = context->inlen;
      } else {
	/*
	  0-------W------------R-------size (inlen = size - R + W)
	  xxxxxxxx             ========
	*/
	available_atonce = context->size - context->rpos;
      }
      if(available_atonce > remain) available_atonce = remain;
      DEBUG(fprintf(stderr, "available_atonce: %d\n", available_atonce));

      memcpy(destpointer, context->buf + context->rpos, available_atonce);
      context->rpos += available_atonce;
      context->inlen -= available_atonce;

      if(context->rpos >= context->size){
	if(context->rpos != context->size){
	  fprintf(stderr, "rpos violation\n");
	}
	context->rpos = 0;
      }
      result += available_atonce;
      destpointer += available_atonce;
    }
  }
  DEBUG(fprintf(stderr, "ringbuffer_get exit(result = %d)\n", result));
  SHOW_STATUS();
  return result;
}

int ringbuffer_put(ringbuffer_context* context, unsigned char* buf, int size)
{
  int result;
  int remain;
  int available_atonce;
  unsigned char* srcpointer;

  for(srcpointer = buf, result = 0; result < size;){
    DEBUG(fprintf(stderr, "ringbuffer_put entrance(request = %d)\n", size));
    SHOW_STATUS();
    if(context->inlen == context->size){ /* buffer full */
      if((context->blocking_bitmap & RINGBUFFER_BLOCKING_PUT) &&
	 context->delayfunction != NULL){
	DEBUG(fprintf(stderr, "buffer full. waiting...\n"));
	context->delayfunction(RINGBUFFER_WAIT_RESOLUTION);
      } else {
	break;
      }
    } else {
      remain =  size - result;
      if(context->rpos <= context->wpos){
	/*
	  0--------R----------W------size  (inlen = W - R)
                              =======
	  0--------R-----------------size  (inlen = W - R = 0)
	           W                          
		   ==================
	*/
	available_atonce = context->size - context->wpos;
      } else {
	/*
	  0-----W------------R-------size  (inlen = size - R + W)
	        =============
	*/
	available_atonce = context->rpos - context->wpos;
      }
      if(available_atonce > remain) available_atonce = remain;
      DEBUG(fprintf(stderr, "available_atonce: %d\n", available_atonce));

      memcpy(context->buf + context->wpos, srcpointer, available_atonce);

      context->inlen += available_atonce;
      context->wpos += available_atonce;
      if(context->wpos >= context->size){
	if(context->wpos != context->size){
	  fprintf(stderr, "wpos violation.\n");
	}
	context->wpos = 0;
      }
      result += available_atonce;
      srcpointer += available_atonce;
    }
  }
  DEBUG(fprintf(stderr, "ringbuffer_put exit(result = %d)\n", result));
  SHOW_STATUS();
  return result;
}

